/*
    ServicesCtl is a programm to start and stop Services from the TrayBar.
    Copyright (C) 2007  Olaf Groeger

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

    For any feedback contact me at servicesctl@olaf-groeger.de.
*/
using System;
using System.Collections.Generic;
using System.Reflection;
using System.Resources;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Security.Permissions;
using System.Security;
using System.Text;
using System.Windows.Forms;

using log4net;

namespace ServicesCtl
{
    /// <summary>
    /// This class handles the impersonation of the current user to an Administator account.
    /// </summary>
    class Administrator
    {
        [DllImport("advapi32.dll", SetLastError = true)]
        public static extern bool LogonUser(String lpszUsername, String lpszDomain, String lpszPassword,
            int dwLogonType, int dwLogonProvider, ref IntPtr phToken);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto)]
        public extern static bool CloseHandle(IntPtr handle);

        [DllImport("kernel32.dll")]
        private unsafe static extern int FormatMessage(int dwFlags, int lpSource,
            int dwMessageId, int dwLanguageId, ref String lpBuffer, int nSize, int Arguments);

        [DllImport("advapi32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        public static extern int DuplicateToken(IntPtr hToken, int impersonationLevel, ref IntPtr hNewToken);


        /// <summary>
        /// A private ResourceManager for handling internationalization.
        /// </summary>
        private ResourceManager messages = null;
        /// <summary>
        /// The WindowsIdentity of an Administrator or null.
        /// </summary>
        private WindowsIdentity windowsIdentity = null;
        /// <summary>
        /// The WindowsImpersonationContext after impersonation to an Administrator, used to "De"-Impersonate
        /// </summary>
        private WindowsImpersonationContext impersonatedUser = null;
        /// <summary>
        /// The private Logger
        /// </summary>
        private static readonly ILog log = LogManager.GetLogger(typeof(Administrator));

        /// <summary>
        /// Creates a new instance.
        /// </summary>
        /// <remarks>Initializes the ResourceManager Administrator.messages.</remarks>
        public Administrator()
        {
            this.messages = new ResourceManager("ServicesCtl.Messages", Assembly.GetExecutingAssembly());
        }

        /// <summary>
        /// Creates a LoginToken for an administrator.
        /// </summary>
        /// <returns>true when a LoginToken with Administrator rights has been created.</returns>
        /// <remarks>Opens a LoginForm to ask for Login data and tries to login with these data.</remarks>
        private bool login()
        {
            //
            // Check that the current user is an Administrator. When this is true, i set the LoginToken
            // to the token of the current user and can skip the whole Login stuff.
            IntPtr tempLoginToken = WindowsIdentity.GetCurrent().Token;
            if (IsAdministrator(tempLoginToken))
            {
                this.windowsIdentity = WindowsIdentity.GetCurrent();
                return true;
            }
            //
            // Prepare the LoginForm
            LoginForm form = new LoginForm();
            bool loggedIn = false;
            bool isAdmin = false;
            do
            {
                // Show the Login dialog
                DialogResult dlgResult = form.ShowDialog();
                //
                // When the user cancels the dialog, then i skip the start/stop request
                if (dlgResult == DialogResult.Cancel)
                    return false;
                if ((form.LoginName == null) || (form.LoginName.Trim().Length == 0))
                    return false;

                if((form.LoginPassword == null ) || (form.LoginPassword.Trim().Length == 0 ))
                    return false;

                const int LOGON32_PROVIDER_DEFAULT = 0;
                //This parameter causes LogonUser to create a primary token.
                const int LOGON32_LOGON_INTERACTIVE = 2;
                // Call LogonUser to obtain a handle to an access token.
                loggedIn = LogonUser(form.LoginName, form.LoginDomain, form.LoginPassword,
                    LOGON32_LOGON_INTERACTIVE, LOGON32_PROVIDER_DEFAULT,
                    ref tempLoginToken);
                // Succesfully logged in?
                if( false == loggedIn )
                {
                    int lastError = Marshal.GetLastWin32Error();
                    String errorMessage = GetErrorMessage(lastError); 
                    log.Error("Login failed: ErrorCode="
                            + lastError
                            + ", \""
                            + errorMessage
                            + "\"");
                    String errorCaption = this.messages.GetString("Administrator_LoginFailed_Caption");
                    MessageBox.Show(errorMessage, errorCaption, MessageBoxButtons.OK);
                    continue;
                }
                // Is the logged in user really an Administrator?
                isAdmin = IsAdministrator(tempLoginToken);
                if (isAdmin == false)
                {
                    String errorMessage = this.messages.GetString("Administrator_NotAdmin_Text");
                    String errorCaption = this.messages.GetString("Administrator_NotAdmin_Caption");
                    MessageBox.Show(errorMessage,errorCaption,MessageBoxButtons.OK);
                }
            } while( (loggedIn == false) || (isAdmin==false) );

            this.windowsIdentity = CreateWindowsIdentity( tempLoginToken );
            return true;
        }

        /// <summary>
        /// Impersonates the current user to an Administrator.
        /// </summary>
        /// <returns>true when the Thread is now in the context of an Administrator</returns>
        /// <remarks>Creates an Administrator LoginToken when needed.</remarks>
        public bool Impersonate()
        {
            if (this.windowsIdentity == null)
            {
                if( !login() )
                    return false;
            }
            try
            {
                this.impersonatedUser = this.windowsIdentity.Impersonate();
                return true;
            }
            catch (System.InvalidOperationException e)
            {
                log.Error("Impersonate failed", e);
            }
            catch (System.Security.SecurityException e)
            {
                log.Error("Impersonate failed", e);
            }
            return false;
        }

        /// <summary>
        /// Resets the current Thread to the current user.
        /// </summary>
        public void Deimpersonate()
        {
            if (this.impersonatedUser != null)
            {
                this.impersonatedUser.Undo();
                this.impersonatedUser = null;
            }
        }

        /// <summary>
        /// Check whether the user is part of the administrator group
        /// </summary>
        /// <param name="loginToken">the LoginToken to check</param>
        /// <returns>true when the LoginToken represents an Administrator account.</returns>
        private bool IsAdministrator(IntPtr loginToken)
        {
            if (loginToken == IntPtr.Zero)
                return false;

            // AppDomain.CurrentDomain.SetPrincipalPolicy(PrincipalPolicy.WindowsPrincipal);
            WindowsIdentity winIdentity = new WindowsIdentity(loginToken);
            WindowsPrincipal principal = new WindowsPrincipal(winIdentity);

            return principal.IsInRole(WindowsBuiltInRole.Administrator);
        }

        /// <summary>
        /// Creates a WindowsIdentity from a Handle.
        /// </summary>
        /// <param name="loginToken">The Handle to convert</param>
        /// <returns>WindowsIdentity or null</returns>
        private WindowsIdentity CreateWindowsIdentity(IntPtr loginToken)
        {
            // Prepare the Inpersonation.
            // I guess impersonation an Administrator to itself doesn't hurt.
            IntPtr duplicateToken = IntPtr.Zero;
            if (DuplicateToken(loginToken, 2, ref duplicateToken) == 0)
            {
                log.Error("Duplicate Token failed");
                return null;
            }
            try
            {
                WindowsIdentity newId = new WindowsIdentity(duplicateToken);
                return newId;
            }
            catch (ArgumentException aEx)
            {
                log.Error("Error creating WindowsIdentity:", aEx);
            }
            catch (SecurityException sEx)
            {
                log.Error("Error creating WindowsIdentity:", sEx);
            }
            finally
            {
                CloseHandle(duplicateToken);
            }
            return null;
        }

        /// <summary>
        /// Helper method to request an error message for a given error number from the OS.
        /// </summary>
        /// <param name="errorCode">The Error code</param>
        /// <returns>The Error message or null</returns>
        public static string GetErrorMessage(int errorCode)
        {
            int FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100;
            int FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200;
            int FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000;

            int messageSize = 255;
            string lpMsgBuf = "";
            int dwFlags = FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS;

            int retVal = FormatMessage(dwFlags, 0, errorCode, 0, ref lpMsgBuf, messageSize, 0);
            if (0 == retVal)
            {
                return null;
            }
            else
            {
                return lpMsgBuf;
            }
        }
    }
}
